#!/bin/bash
#
# This executable converts a Mavericks .app (which allows to upgrade a machine
# from Mac OS 10.6.7+ to Mavericks) into a Mavericks .vmdk (which allows to
# install Mavericks from scratch in a virtual machine).
#
# It has been tested with the following Mac OS versions
# o 10.9 (starting with Developer Preview 4)
# o 10.10
# o 10.11
# o 10.12
# o 10.13
# o 10.14
# o 10.15
#
# Note: "IA" below does not stand for Intel Architecture, it stands for Install
# Assistant.
#

# Uncomment the following line to get debug output.
#exec >/tmp/debug 2>&1

set -x
set -e
set -E

# The first argument is the path to the .app bundle (the input of the
# executable).
inputApp="$1"
inputAppB="`basename "$inputApp"`"
inputAppU="${inputAppB// /%20}"

# The second argument is the path to the .vmdk file (the output of the
# executable), which must end with ".vmdk".
outputVmdk="$2"
[ "${outputVmdk: -5}" = .vmdk ]
outputDmg="${outputVmdk:0:${#outputVmdk} - 5}".dmg

# The macOS installer fails cryptically (the "Continue" button in macOS
# Utilities will do nothing other than disable itself) if the contained
# framework symlinks are no longer in place, e.g. replaced with the referenced
# content.  This may occur if the bundle has been transferred through a medium
# or protocol which does not preserve symlinks.  Fail early if we detect this.
if find "$inputApp" -path '*.framework/Versions/Current' -type d | \
   grep -m1 . ; then
   echo "Framework symlinks missing.  Source image is corrupt." >&2
   exit 1
fi

# createinstallmedia hardcodes the volume name. The installation succeeds
# regardless of the name, so let us just pick something sensible.
volName="`defaults read "$inputApp"/Contents/Info CFBundleDisplayName`"

tmpDir="`mktemp -d -t 'Create Mavericks Installer'`"
outputMnt="$tmpDir"/output
installMnt="$tmpDir"/install
baseMnt="$tmpDir"/base

# Detach a disk image.  Another process might be using the disk (i.e. a malware
# scanner), so we perform a few retries, and add "-force" if we still can't
# detach it.
hdiutilDetachWithRetries() {
   local interval=1
   while ! hdiutil detach "$@" ; do
      if [ $interval -gt 16 ] ; then
         hdiutil detach -force "$@" || echo "Unable to detach disk." >&2
         break
      fi
      sleep $interval
      interval=$((interval * 2))
   done
}

cleanup() {
   local status="$?"

   if [ -d "$outputMnt" ]; then
      hdiutilDetachWithRetries "$outputMnt"
   fi

   if [ -d "$baseMnt" ]; then
      hdiutilDetachWithRetries "$baseMnt"
   fi

   if [ -d "$installMnt" ]; then
      hdiutilDetachWithRetries "$installMnt"
   fi

   rm -rf -- "$tmpDir"

   if [ "$status" -ne 0 ]; then
      rm -f "$outputVmdk" "$outputDmg"
   fi
}

# Cleanup on failure.
trap cleanup ERR

# Mount InstallESD.dmg so we can access BaseSystem.dmg inside.
# This fails if the image is already mounted somewhere else.
# So force to deatch it first.
diskID=`hdiutil \
   attach "$inputApp"/Contents/SharedSupport/InstallESD.dmg \
   -mount suppressed \
   -noverify \
   | cut -d' ' -f 1 | head -n 1`
hdiutil detach -force $diskID || echo "Unable to detach disk." >&2
hdiutil \
   attach "$inputApp"/Contents/SharedSupport/InstallESD.dmg \
   -mountpoint "$installMnt" \
   -nobrowse \
   -noverify \
   -readonly

#
# Determine the location of the BaseSystem.dmg.
#
# Newer than Mac OS 10.12:
baseSystem="$inputApp"/Contents/SharedSupport/BaseSystem.dmg
if [ ! -r "$baseSystem" ]; then
   # Mac OS 10.12 or earlier:
   baseSystem="$installMnt"/BaseSystem.dmg
fi

# Mount BaseSystem.dmg so we can access files inside.
# This fails if the image is already mounted somewhere else.
hdiutil \
   attach "$baseSystem" \
   -mountpoint "$baseMnt" \
   -nobrowse \
   -noverify \
   -readonly

#
# Determine the size of the disk image.
#
# Between the initial boot and the first reboot, the Mavericks installer writes
# 16.2 MB to the volume, mostly to replace the
#    .IABootFiles/boot.efi
#    .IABootFiles/"$prelinkedKernelB"
#    .IABootFiles/PlatformSupport.plist
#    .IABootFiles/com.apple.Boot.plist
# hard links with actual copies...
#
# So account for that. To err on the side of caution, we further enlarge the
# disk image to keep 30 MB of free space on the volume at the first reboot.
#

addSize() {
   # Use a temp file instead of a shell pipe, to detect when "du" fails.
   BLOCKSIZE=512 du -s "$2" >"$tmpDir"/pipe
   local size="`awk '{ print $1 }' "$tmpDir"/pipe`"
   outputDmgSize=$((outputDmgSize + $1 * size))
}

#
# Determine the location of the prelinked kernel (a.k.a. kernelcache).
#
# There are two possible locations for the prelinked kernel.  The first
# location is used by newer products.  If not found there, fall back to the
# location used by Mac OS versions 10.9 and 10.10.
#

# Newer than Mac OS 10.10:
prelinkedKernel=System/Library/PrelinkedKernels/prelinkedkernel
if [ ! -r "$baseMnt"/"$prelinkedKernel" ]; then
   # Mac OS 10.9 or 10.10:
   prelinkedKernel=System/Library/Caches/com.apple.kext.caches/Startup/kernelcache
fi
prelinkedKernelB="`basename "$prelinkedKernel"`"

#
# Determine the target location of the prelinked kernel.
#
if [ -d "$baseMnt"/System/Library/Caches ]; then
   # Mac OS 10.11 or earlier:
   prelinkedKernelT=System/Library/Caches/com.apple.kext.caches/Startup/"$prelinkedKernelB"
else
   # Newer than Mac OS 10.11:
   prelinkedKernelT=System/Library/PrelinkedKernels/"$prelinkedKernelB"
fi

# A new booter System/Library/CoreServices/bootbase.efi was introduced in
# Mac OS 10.11 Public Beta 1. Apple recommends we use it going forward.
booter=System/Library/CoreServices/bootbase.efi
if [ ! -r "$baseMnt"/"$booter" ]; then
   booter=System/Library/CoreServices/boot.efi
fi

outputDmgSize=119200
addSize 1 "$inputApp"
for i in \
   "$prelinkedKernel" \
   System/Library/CoreServices/PlatformSupport.plist \
   "$booter" \
   ; do
   addSize 2 "$baseMnt"/"$i"
done

# Create the disk image.
rm -f "$outputDmg"
# Layout: MBRSPUD produces a smaller disk image than SPUD or GPTSPUD.
hdiutil \
   create "$outputDmg" \
   -sectors "$outputDmgSize" \
   -fs HFS+J \
   -volname "$volName" \
   -layout MBRSPUD

# Mount the disk image.
hdiutil \
   attach "$outputDmg" \
   -mountpoint "$outputMnt" \
   -nobrowse \
   -noverify

#
# Copy some files.  The majority of the work (copying the installer .app bundle
# itself) takes place later.
#

copy() {
   mkdir -p "`dirname "$2"`"
   cp -a "$1" "$2"
}

copyRel() {
   copy "$baseMnt"/"$1" "$outputMnt"/"$2"
}

copyRel "$prelinkedKernel" \
        "$prelinkedKernelT"
copyRel System/Library/CoreServices/PlatformSupport.plist \
        System/Library/CoreServices/PlatformSupport.plist
copyRel "$booter" \
        System/Library/CoreServices/boot.efi
copyRel System/Library/CoreServices/SystemVersion.plist \
        System/Library/CoreServices/SystemVersion.plist
# A new volume icon was introduced in Mac OS 10.13 Public Beta 1.
if [ -r "$inputApp"/Contents/SharedSupport/BaseSystem.dmg ]; then
   copy "$inputApp"/Contents/Resources/ProductPageIcon.icns \
        "$outputMnt"/.VolumeIcon.icns
fi

#
# Determine the kernel flags used in com.apple.Boot.plist.
#
if [ -r "$inputApp"/Contents/SharedSupport/BaseSystem.dmg ]; then
   # Newer than Mac OS 10.12:
   kernelFlags="root-dmg=file:///"$inputAppU"/Contents/SharedSupport/BaseSystem.dmg"
else
   # Mac OS 10.12 or earlier:
   kernelFlags="container-dmg=file:///"$inputAppU"/Contents/SharedSupport/InstallESD.dmg root-dmg=file:///BaseSystem.dmg"
fi

#
# Generate some files.
#

f="$outputMnt"/Library/Preferences/SystemConfiguration/com.apple.Boot.plist
mkdir -p "`dirname "$f"`"
cat <<EOF >"$f"
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
	<key>Kernel Cache</key>
	<string>/.IABootFiles/$prelinkedKernelB</string>
	<key>Kernel Flags</key>
	<string>$kernelFlags</string>
</dict>
</plist>
EOF

f="$outputMnt"/.IAPhysicalMedia
mkdir -p "`dirname "$f"`"
cat <<EOF >"$f"
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
	<key>AppName</key>
	<string>$inputAppB</string>
</dict>
</plist>
EOF

#
# Hardlink some files.
#

hardlink() {
   mkdir -p "`dirname "$outputMnt"/"$2"`"

   # "ln" fails with "Operation not permitted" (even for the super user) if the
   # source's uchg (user immutable) flag is set.

   local flags=`ls -lO@ "$outputMnt"/"$1" | awk '{ print $5 }'`

   if echo "$flags" | grep -q uchg; then
      # The uchg flag is set. Temporarily clear it to allow "ln" to succeed.
      chflags nouchg "$outputMnt"/"$1"
   fi

   ln "$outputMnt"/"$1" "$outputMnt"/"$2"

   if echo "$flags" | grep -q uchg; then
      # The uchg flag was set. Restore it.
      chflags uchg "$outputMnt"/"$1"
   fi
}

hardlink System/Library/CoreServices/boot.efi \
   usr/standalone/i386/boot.efi
hardlink System/Library/CoreServices/boot.efi \
   .IABootFiles/boot.efi
hardlink "$prelinkedKernelT" \
   .IABootFiles/"$prelinkedKernelB"
hardlink System/Library/CoreServices/PlatformSupport.plist \
   .IABootFiles/PlatformSupport.plist
hardlink Library/Preferences/SystemConfiguration/com.apple.Boot.plist \
   .IABootFiles/com.apple.Boot.plist
# The target path should probably be .IABootFiles/SystemVersion.plist, but let
# us be bug-compatible with Apple :)
hardlink System/Library/CoreServices/SystemVersion.plist \
   .IABootFilesSystemVersion.plist

#
# Copy the bulky installer .app bundle.
# The copy must preserve symbolic links (see bug 1677798).
#

cp -a "$inputApp" "$outputMnt"

# Bless the volume.
/usr/sbin/bless \
   -folder "$outputMnt"/.IABootFiles \
   -bootefi "$outputMnt"/.IABootFiles/boot.efi \
   -options 'config="\.IABootFiles\com.apple.Boot"' \
   -label "$volName"

# Delete unnecessary stuff that might have been created by Finder.
if [ -e "$outputMnt"/.Trashes ]; then
   chmod u+r "$outputMnt"/.Trashes
fi
rm -rf "$outputMnt"/.{DS_Store,Spotlight-V100,Trashes,fseventsd}

# Synchronize "$outputDmg".
hdiutilDetachWithRetries "$outputMnt"

# Create the virtual disk descriptor.
cat <<EOF >"$outputVmdk"
# Disk DescriptorFile
version=1
encoding="UTF-8"
CID=fffffffe
parentCID=ffffffff
isNativeSnapshot="no"
createType="monolithicFlat"

# Extent description
RW $outputDmgSize FLAT "`basename "$outputDmg"`" 0

# The Disk Data Base 
#DDB

ddb.adapterType = "lsilogic"
#ddb.geometry.cylinders is not used by Mac OS.
#ddb.geometry.heads is not used by Mac OS.
#ddb.geometry.sectors is not used by Mac OS.
#ddb.longContentID will be generated on the first write to the file.
#ddb.uuid is not used by Mac OS.
ddb.virtualHWVersion = "6"
EOF

# Cleanup on success.
trap ERR; cleanup
