#!/bin/bash
#
# This executable converts an Install macOS.app into a VMware virtual disk
# (.vmdk), which allows installation of the OS from scratch in a virtual
# machine).
#
# It has been tested with the following Mac OS versions
# o 11.0 Beta 1 through 6, Beta 9
#

# Uncomment the following line to get debug output.
#exec >/tmp/debug 2>&1

set -x
set -e
set -E

# The first argument is the path to the .app bundle (the input of the
# executable).
inputApp="$1"
inputAppB="`basename "$inputApp"`"

if [ \! -r "$inputApp"/Contents/SharedSupport/SharedSupport.dmg ] ; then
   # Pass control to the script for earlier macOS releases.
   exec "`dirname "$0"`/Create Mavericks Installer.tool" "$@"
fi

# The second argument is the path to the .vmdk file (the output of the
# executable), which must end with ".vmdk".
outputVmdk="$2"
[ "${outputVmdk: -5}" = .vmdk ]
outputDmg="${outputVmdk:0:${#outputVmdk} - 5}".dmg

# The macOS installer fails cryptically (the "Continue" button in macOS
# Utilities will do nothing other than disable itself) if the contained
# framework symlinks are no longer in place, e.g. replaced with the referenced
# content.  This may occur if the bundle has been transferred through a medium
# or protocol which does not preserve symlinks.  Fail early if we detect this.
if find "$inputApp" -path '*.framework/Versions/Current' -type d | \
   grep -m1 . ; then
   echo "Framework symlinks missing.  Source image is corrupt." >&2
   exit 1
fi

# createinstallmedia hardcodes the volume name. The installation succeeds
# regardless of the name, so let us just pick something sensible.
volName="`defaults read "$inputApp"/Contents/Info CFBundleDisplayName`"

tmpDir="`mktemp -d -t 'Create macOS Installer'`"
outputMnt="$tmpDir"/output
installMnt="$tmpDir"/install
baseMnt="$tmpDir"/base
baseDiskList="$tmpDir"/base-disks.txt

# Detach a disk image.  Another process might be using the disk (i.e. a malware
# scanner), so we perform a few retries, and add "-force" if we still can't
# detach it.
hdiutilDetachWithRetries() {
   local interval=1
   while ! hdiutil detach "$@" ; do
      if [ $interval -gt 16 ] ; then
         hdiutil detach -force "$@" || echo "Unable to detach disk." >&2
         break
      fi
      sleep $interval
      interval=$((interval * 2))
   done
}

detachBaseDisk() {
   if [ -r "$baseDiskList" ]; then
      # Detach the first disk device mounted by us.
      local baseDev

      baseDev=`cut -f1 "$baseDiskList" | \
               sed -ne '/^\(\/dev\/disk[0-9]*\)[[:blank:]]*$/ {s//\1/;p;q;}'`
      if [ -e "$baseDev" ] ; then
         hdiutilDetachWithRetries "$baseDev"
      fi
   fi
}

cleanup() {
   local status="$?"

   if [ -d "$outputMnt" ]; then
      hdiutilDetachWithRetries "$outputMnt"
   fi

   detachBaseDisk

   if [ -d "$installMnt" ]; then
      hdiutilDetachWithRetries "$installMnt"
   fi

   rm -rf -- "$tmpDir"

   if [ "$status" -ne 0 ]; then
      rm -f "$outputVmdk" "$outputDmg"
   fi
}

# Cleanup on failure.
trap cleanup ERR

# Mount SharedSupport.dmg so we can access the assets inside.
# This fails if the image is already mounted somewhere else,
# so force to deatch it first.
diskID=`hdiutil \
   attach "$inputApp"/Contents/SharedSupport/SharedSupport.dmg \
   -mount suppressed \
   -noverify \
   | cut -d' ' -f 1 | head -n 1`
hdiutil detach -force $diskID || echo "Unable to detach disk." >&2
hdiutil \
   attach "$inputApp"/Contents/SharedSupport/SharedSupport.dmg \
   -mountpoint "$installMnt" \
   -nobrowse \
   -noverify \
   -readonly

assetXmlPath="$installMnt"/com_apple_MobileAsset_MacSoftwareUpdate/\
com_apple_MobileAsset_MacSoftwareUpdate.xml

# Obtain the path to the first installation asset.
assetPath=`plutil -extract 'Assets.0.__RelativePath' xml1 "$assetXmlPath" \
               -o - | \
           plutil -p - | \
           sed -e 's/^"\(.*\)"/\1/'`

addSize() {
   # Use a temp file instead of a shell pipe, to detect when "du" fails.
   BLOCKSIZE=512 du -s "$1" >"$tmpDir"/pipe
   local size="`awk '{ print $1 }' "$tmpDir"/pipe`"
   outputDmgSize=$((outputDmgSize + size))
}

assetSize() {
   unzip -qql "$installMnt"/"$assetPath" "$@" | \
   awk '
BEGIN {SUM=0}
{
   SUM = SUM + int(($1 + 4095) / 4096)
}
END {print SUM * 8}'
}

addAssetSize() {
   outputDmgSize=$((outputDmgSize + `assetSize "$@"`))
}

outputDmgSize=119200
# Add an extra 300 MBytes due to bug 2623782.
outputDmgSize=$((outputDmgSize + 300 * 1024 * 1024 / 512))
addSize "$inputApp"
addAssetSize AssetData/boot/Firmware/Manifests/InstallerBoot/'*' \
             AssetData/Restore/'*' \
             AssetData/boot/BuildManifest.plist

# We have a bit of a challenge here.  Some other files need to be copied into
# the output disk from within BaseSystem.dmg, but that's inside a zipfile and
# we can't fetch the sizes until the zipfile is decompressed, and it would be
# rather wasteful to uncompress the zipfile just to query the sizes.
#
# They are (from inside BaseSystem.dmg):
#
#    System/Library/CoreServices/boot.efi
#    System/Library/CoreServices/BridgeVersion.bin
#    System/Library/CoreServices/PlatformSupport.plist
#    System/Library/CoreServices/SystemVersion.plist
#    System/Library/KernelCollections/BootKernelExtensions.kc
#    System/Library/PrelinkedKernels/prelinkedkernel (optional)
#    usr/standalone/i386/SecureBoot.bundle/
#
# Currently, for macOS 11 Beta 1, they total 92 MBytes.  Allow 200 MBytes.
outputDmgSize=$((outputDmgSize + 200 * 1024 * 1024 / 512))

# Create the disk image.
rm -f "$outputDmg"
# Layout: MBRSPUD produces a smaller disk image than SPUD or GPTSPUD.
hdiutil \
   create "$outputDmg" \
   -sectors "$outputDmgSize" \
   -fs HFS+J \
   -volname "$volName" \
   -layout MBRSPUD

# Mount the disk image.
hdiutil \
   attach "$outputDmg" \
   -mountpoint "$outputMnt" \
   -nobrowse \
   -noverify

touch "$outputMnt"/.metadata_never_index

#
# Copy some files.  The majority of the work (copying the installer .app bundle
# itself) takes place later.
#

copy() {
   mkdir -p "`dirname "$2"`"
   cp -a "$1" "$2"
}

copyRel() {
   copy "$baseSystemMnt"/"$1" "$outputMnt"/"$2"
}

unzipRemovingPrefix() {
   # Unzips "$1"/"$2" to "$3"/"$2".
   #
   # The "unzip" program doesn't support stripping some (but not all) leading
   # path components.  Use a symlink to achieve the desired effect.
   local prefixDir="$tmpDir"/prefix

   mkdir -p "$prefixDir"/"`dirname "$1"`"
   mkdir -p "$outputMnt"/"$3"
   ln -s "$outputMnt"/"$3" "$prefixDir"/"$1"
   unzip -d "$prefixDir" "$installMnt"/"$assetPath" "$1"/"$2"
   rm -fR "$prefixDir"
}

unzipRemovingPrefix AssetData/boot/Firmware/Manifests/InstallerBoot '*' /
unzipRemovingPrefix AssetData/Restore '*' /BaseSystem

baseSystem="$outputMnt"/BaseSystem/BaseSystem.dmg

# Mount BaseSystem.dmg so we can access files inside.
# This fails if the image is already mounted somewhere else.
# The image contains an APFS container with several APFS volumes, so hdiutil
# will typically create at least two device nodes (one for the image, one for
# the APFS container) and three slices (one for the APFS container slice inside
# the image, and two for the APFS volumes inside the container).  hdiutil's
# output will tell us where we can find the mounted volume(s) as well as the
# identity of the device nodes.
hdiutil \
   attach "$baseSystem" \
   -mountroot "$baseMnt" \
   -nobrowse \
   -noverify \
   -readonly > "$baseDiskList"

baseSystemMnt="`cut -f3 "$baseDiskList" | grep -m1 'Base System'`"

for f in System/Library/CoreServices/boot.efi \
         System/Library/CoreServices/BridgeVersion.bin \
         System/Library/CoreServices/PlatformSupport.plist \
         System/Library/CoreServices/SystemVersion.plist \
         System/Library/KernelCollections/BootKernelExtensions.kc \
         usr/standalone/i386/SecureBoot.bundle ; do
   copyRel "$f" "$f"
done

prelinkedKernel="System/Library/PrelinkedKernels/prelinkedkernel"
if [ -e "$baseSystemMnt"/"$prelinkedKernel" ] ; then
   copyRel "$prelinkedKernel" "$prelinkedKernel"
else
   unset prelinkedKernel
fi

detachBaseDisk

unzipRemovingPrefix AssetData/boot BuildManifest.plist \
   /usr/standalone/i386/SecureBoot.bundle/Contents/Resources

copy "$inputApp"/Contents/Resources/ProductPageIcon.icns \
     "$outputMnt"/.VolumeIcon.icns

#
# Generate some files.
#

f="$outputMnt"/Library/Preferences/SystemConfiguration/com.apple.Boot.plist
mkdir -p "`dirname "$f"`"
cat <<EOF >"$f"
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
	<key>Kernel Flags</key>
	<string>root-dmg=file:///BaseSystem/BaseSystem.dmg</string>
</dict>
</plist>
EOF

f="$outputMnt"/.IAPhysicalMedia
mkdir -p "`dirname "$f"`"
cat <<EOF >"$f"
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
	<key>AppName</key>
	<string>$inputAppB</string>
</dict>
</plist>
EOF

#
# Hardlink some files.
#

hardlink() {
   mkdir -p "`dirname "$outputMnt"/"$2"`"

   # "ln" fails with "Operation not permitted" (even for the super user) if the
   # source's uchg (user immutable) flag is set.

   local flags=`ls -lO@ "$outputMnt"/"$1" | awk '{ print $5 }'`

   if echo "$flags" | grep -q uchg; then
      # The uchg flag is set. Temporarily clear it to allow "ln" to succeed.
      chflags nouchg "$outputMnt"/"$1"
   fi

   ln "$outputMnt"/"$1" "$outputMnt"/"$2"

   if echo "$flags" | grep -q uchg; then
      # The uchg flag was set. Restore it.
      chflags uchg "$outputMnt"/"$1"
   fi
}

if [ -n "$prelinkedKernel" ] ; then
   hardlink "$prelinkedKernel" System/Library/PrelinkedKernels/immutablekernel
fi

#
# Copy the bulky installer .app bundle.
# The copy must preserve symbolic links (see bug 1677798).
#

cp -a "$inputApp" "$outputMnt"

# Bless the volume.
/usr/sbin/bless \
   -folder "$outputMnt"/System/Library/CoreServices \
   -bootefi "$outputMnt"/System/Library/CoreServices/boot.efi \
   -label "$volName"

# Delete unnecessary stuff that might have been created by Finder.
if [ -e "$outputMnt"/.Trashes ]; then
   chmod u+r "$outputMnt"/.Trashes
fi
rm -rf "$outputMnt"/.{DS_Store,Spotlight-V100,Trashes,fseventsd}

# Synchronize "$outputDmg".
hdiutilDetachWithRetries "$outputMnt"

# Create the virtual disk descriptor.
cat <<EOF >"$outputVmdk"
# Disk DescriptorFile
version=1
encoding="UTF-8"
CID=fffffffe
parentCID=ffffffff
isNativeSnapshot="no"
createType="monolithicFlat"

# Extent description
RW $outputDmgSize FLAT "`basename "$outputDmg"`" 0

# The Disk Data Base 
#DDB

ddb.adapterType = "lsilogic"
#ddb.geometry.cylinders is not used by Mac OS.
#ddb.geometry.heads is not used by Mac OS.
#ddb.geometry.sectors is not used by Mac OS.
#ddb.longContentID will be generated on the first write to the file.
#ddb.uuid is not used by Mac OS.
ddb.virtualHWVersion = "6"
EOF

# Cleanup on success.
trap ERR; cleanup
