#!/bin/bash -p
#
# This executable encapsulates all the code which be must run as root once the
# application has been copied to a new location. It is written as a script
# so invokers can more easily see what it does. It is designed to be invoked in
# two ways:
#
# 1) Interactively, by a home user running the application.
#
# When the application starts, it invokes this executable as the current user
# with the first argument set to "test", to determine whether something needs
# to be done as root (non-zero exit status) or not (zero exit status). In case
# something needs to be done, the application prompts the user for root
# credentials, then invokes this executable again as root with the first
# argument set to "set", to actually do what needs to be done as root.
#
# 2) Unattended, by the system administrator of an organization who mass
#    deploys the application.
#
# After the application has been copied to a new machine, the system
# administrator invokes this executable as root with the first argument set to
# "set". After that, any user of the organization who starts the application on
# this machine will not be prompted for root credentials.
#


# Print a variable, even if its value is -n (unlike echo).
print_var() {
   printf -- %s "$1"
}


# Canonicalize the path of an existing directory.
canonicalize_dir() {
   # Bash's cd builtin can write to stdout.
   (cd -- "$1" >/dev/null && pwd -P)
}


# Canonicalize the path of an existing file.
canonicalize_file() {
   local file="$1"

   while [ -h "$file" ]; do
      file="`readlink -- "$file"`"
   done

   local dir
   # Keep the "local" declaration separate from command substitution. PR 1882121.
   dir="`dirname -- "$file"`"

   # 'dir' is a directory, or a symlink to a directory. Canonicalize it.
   local canonicalFile
   # Keep the "local" declaration separate from command substitution. PR 1882121.
   canonicalFile="`canonicalize_dir "$dir"`"
   # Keep the above and below command substitutions separate. PR 1882121.
   canonicalFile="$canonicalFile"/"`basename -- "$file"`"
   # Keep the above and below command substitutions separate. PR 1882121.
   print_var "$canonicalFile"
}


# Test if this executable can do things as root.
canSet() {
   [ "$action" = set -a "$EUID" = 0 ]
}


# Convert a numerical access permission into a string.
accessNum2Str() {
   local result=''

   if [ "$((0$1 & 0400))" != 0 ]; then
      result="$result"r
   else
      result="$result"-
   fi
   if [ "$((0$1 & 0200))" != 0 ]; then
      result="$result"w
   else
      result="$result"-
   fi
   if [ "$((0$1 & 0100))" != 0 ]; then
      if [ "$((0$1 & 04000))" != 0 ]; then
         result="$result"s
      else
         result="$result"x
      fi
   else
      result="$result"-
   fi

   if [ "$((0$1 & 0040))" != 0 ]; then
      result="$result"r
   else
      result="$result"-
   fi
   if [ "$((0$1 & 0020))" != 0 ]; then
      result="$result"w
   else
      result="$result"-
   fi
   if [ "$((0$1 & 0010))" != 0 ]; then
      result="$result"x
   else
      result="$result"-
   fi

   if [ "$((0$1 & 0004))" != 0 ]; then
      result="$result"r
   else
      result="$result"-
   fi
   if [ "$((0$1 & 0002))" != 0 ]; then
      result="$result"w
   else
      result="$result"-
   fi
   if [ "$((0$1 & 0001))" != 0 ]; then
      result="$result"x
   else
      result="$result"-
   fi

   print_var "$result"
}


# Test the Type, Permission and Ownership of a path.
hasTPO() {
   local path="$1"
   local accessStr="$2"
   local uid="$3"
   local gid="$4"

   # Numeric IDs are used because some systems have two names (or more)
   # associated with one number.
   set -- `ls -alnd -- "$path" 2>/dev/null`
   [ "${1:0:10}" = "$accessStr" -a "$3" = "$uid" -a "$4" = "$gid" ]
}


# Test and set existence, permission and ownership of a directory.
checkDir() {
   local path="$1"
   local accessNum="$2"
   local gid="$3"
   local gName="$4"

   if ! hasTPO "$path" d"`accessNum2Str "$accessNum"`" 0 "$gid"; then
      canSet

      if ! [ -d "$path" ]; then
         echo '*** Creating directory '"$path"'...'
         mkdir -- "$path"
      fi

      echo '*** Setting permissions on '"$path"'...'
      chown -- root:"$gName" "$path"
      chmod -- "$accessNum" "$path"
   fi
}


# Verify dir is not empty nor root dir. PR 1882121
verifyDir() {
   [ "$1" != "/" -a "$1" != "" ]
}


# Set the location variables.
locationInit() {
   # PR 1882121
   # set -e doesn't propagate into a command substitution environment.
   # Add set -e in command substitution environment.
   this="`set -e && canonicalize_file "$1"`"
   libDir="`dirname -- "$this"`"
   local contentsDir
   # Keep the "local" declaration separate from command substitution. PR 1882121.
   contentsDir="`dirname -- "$libDir"`"
   appDir="`dirname -- "$contentsDir"`"
}


#
# Main
#

# To debug, uncomment the next 2 lines.
#set -x
#exec >>/tmp/debug 2>&1

set -e
# This executable runs with root privileges, so hardcode a PATH where
# unprivileged users cannot write.
export PATH=/bin:/sbin:/usr/bin:/usr/sbin

# Force the rest of this executable to explicitly set all file permissions.
umask 777

LEGACY_HOST_DIR='/Library/Application Support/VMware Fusion'
HOST_DIR='/Library/Preferences/VMware Fusion'
LOCATION="$HOST_DIR"/lastLocationUsed
action="$1"
guiPid="$2"
newAppDir="$3"
licenseKey="$4"

if [ "$guiPid" != '' ]; then
   # When launched from GUI, Send stderr to stdout and be verbose.
   exec 2>&1
   set -x
fi

locationInit "$0"

if canSet && [ "$newAppDir" != '' ]; then
   rm -rf -- "$newAppDir"
   cp -pR -- "$appDir" "$newAppDir"
   locationInit "$newAppDir"'/Contents/Library/Initialize VMware Fusion.tool'
fi

# Validating appDir.
verifyDir "$appDir"

if ! [ -d "$HOST_DIR" ]; then
   canSet

   echo '*** Creating the host directory...'
   mkdir -m 755 -- "$HOST_DIR"

   echo '*** Copying files from the legacy host directory' \
        '(before version 4rc)...'
   files=''
   files="$files"' config'                         # Versions 1 to 4b1
   files="$files"' license.fusion.site.6.0.200610' # Version 2
   files="$files"' license-fusion-30-e1-200902'    # Version 3
   files="$files"' license-fusion-40-e1-201010'    # Version 4b1
   files="$files"' locations'                      # Versions 1 to 2
   files="$files"' networking'                     # Versions 3 to 4b1
   files="$files"' rawdiskAuthorization'           # Versions 3.1 to 4b1
   for file in $files; do
      if [ -f "$LEGACY_HOST_DIR"/"$file" ]; then
         cp -p -- "$LEGACY_HOST_DIR"/"$file" "$HOST_DIR"
      fi
   done

   for dir in vmnet1 vmnet8; do                    # Versions 1 to 4b1
      if [ -d "$LEGACY_HOST_DIR"/"$dir" ]; then
         cp -pR -- "$LEGACY_HOST_DIR"/"$dir" "$HOST_DIR"
      fi
   done
fi


# See https://reviewboard.eng.vmware.com/r/326779 for details.
SUPPORT_DIR='/Library/Application Support/VMware'
checkDir "$SUPPORT_DIR" 755 0 wheel
checkDir "$SUPPORT_DIR"'/VMware Fusion' 755 0 wheel
checkDir "$SUPPORT_DIR"'/VMware Fusion/AdminWritable' 775 80 admin
checkDir "$SUPPORT_DIR"'/VMware Fusion/Shared' 777 80 admin


LEGACY_UNINSTALLER="$LEGACY_HOST_DIR"'/Uninstall VMware Fusion.app/Contents/Resources/fusion-uninstaller.tool'
if [ -x "$LEGACY_UNINSTALLER" ]; then
   canSet

   echo '*** Uninstalling legacy product (before version 4rc)...'
   # Move ourselves out of reach of the legacy uninstaller in case we happen to
   # be in a directory it is about to delete.
   tmpAppDir="`mktemp -d -t fusion`"
   # Keep the above and below command substitutions separate. PR 1882121.
   tmpAppDir="$tmpAppDir"/"`basename -- "$appDir"`"
   mv -- "$appDir" "$tmpAppDir"
   (umask 022 && "$LEGACY_UNINSTALLER")
   status="$?"
   mv -- "$tmpAppDir" "$appDir"
   rmdir -- "`dirname -- "$tmpAppDir"`"
   [ "$status" -eq 0 ]
fi


# Before you change gName/gid/access below, read comment 11 of bug 1085325.
gName=wheel
gid=0
access=644
xAccess="`printf -- %o "$((0$access | 0111))"`"
sxAccess="`printf -- %o "$((0$xAccess | 04000))"`"
sxAccessStr=-"`accessNum2Str "$sxAccess"`"
# Test if a path is a set-UID root file.
isSetuidRoot() {
   hasTPO "$1" "$sxAccessStr" 0 "$gid"
}
setPerms() {
   local path

   echo '*** Setting file permissions...'

   # First, the default rule.
   # Building the product as an unprivileged user requires write permission to
   # all directories inside the build tree, to be able to update files.
   if ! [ -e "$appDir"'/Contents/Resources/Inside Build Tree' ]; then
      # All files are owned by root:"$gName".
      chown -R -- root:"$gName" "$appDir"
   fi
   # All files have permissions "$access".
   chmod -R -- "$access" "$appDir"
   # All directories have permission "$xAccess".
   find -f "$appDir" -- -type d -exec chmod -- "$xAccess" {} +

   # Then mark some files as executable.
   # The substitution of MAIN_EXECS uses "$appDir".
   execs=("$appDir""/Contents/MacOS/VMware Fusion" "$appDir""/Contents/Library/relaunch.sh" "$appDir""/Contents/Library/vmware-vdiskmanager" "$appDir""/Contents/Library/vmware-cloneBootCamp" "$appDir""/Contents/Library/vmware-rawdiskAuthTool" "$appDir""/Contents/Library/vmware-ntfs" "$appDir""/Contents/Library/unrar" "$appDir""/Contents/Library/ntfsls" "$appDir""/Contents/Library/ntfscat" "$appDir""/Contents/Library/mkntfs" "$appDir""/Contents/Library/ntfsresize" "$appDir""/Contents/Library/ntfscp" "$appDir""/Contents/Library/shares/fusionShareTool" "$appDir""/Contents/Library/shares/adduser.sh" "$appDir""/Contents/Library/shares/addshare.sh" "$appDir""/Contents/Library/shares/rmuser.sh" "$appDir""/Contents/Library/shares/rmshare.sh" "$appDir""/Contents/Library/dockers/fusionRepairPermissionsTool" "$appDir""/Contents/Library/dockers/proxyApp-template-app/Contents/MacOS/vmware-docker" "$appDir""/Contents/Library/vmware-usbarbitrator" "$appDir""/Contents/Library/services/Open VMware USB Arbitrator Service" "$appDir""/Contents/Library/services/VMware USB Arbitrator Service" "$appDir""/Contents/Library/services/usbarbService.sh" "$appDir""/Contents/Library/services/VMware Fusion Services" "$appDir""/Contents/Library/services/services.sh" "$appDir""/Contents/Library/services/Open VMware Graphics Service" "$appDir""/Contents/Library/services/VMware Graphics Service" "$appDir""/Contents/Library/vmware-aewp" "$appDir""/Contents/Library/scriptreg" "$appDir""/Contents/Library/regval" "$appDir""/Contents/Library/vmware-id" "$appDir""/Contents/Library/Create macOS Installer.tool" "$appDir""/Contents/Library/Create Mavericks Installer.tool" "$appDir""/Contents/Library/vmss2core" "$appDir""/Contents/Library/vmware-vmdkserver" "$appDir""/Contents/Library/VMware Fusion Problem Reporter.tool" "$appDir""/Contents/Library/vmnet-bridge" "$appDir""/Contents/Library/vmnet-natd" "$appDir""/Contents/Library/vmnet-dhcpd" "$appDir""/Contents/Library/vmnet-sniffer" "$appDir""/Contents/Library/vmnet-netifup" "$appDir""/Contents/Library/vmnet-cfgcli" "$appDir""/Contents/Library/vmnet-cli" "$appDir""/Contents/Library/vmrun" "$appDir""/Contents/Library/licenses/vmware-licenseTool" "$appDir""/Contents/Library/VMware Fusion Start Menu.app/Contents/MacOS/VMware Fusion Start Menu" "$appDir""/Contents/Library/vmware-remotemks" "$appDir""/Contents/Library/mksSandbox" "$appDir""/Contents/Library/mksSandbox-debug" "$appDir""/Contents/Library/mksSandbox-stats" "$appDir""/Contents/Library/LaunchServices/com.vmware.KextControlHelper" "$appDir""/Contents/Library/LaunchServices/com.vmware.MountHelper" "$appDir""/Contents/Library/LaunchServices/com.vmware.VMMonHelper" "$appDir""/Contents/Library/LaunchServices/com.vmware.DiskHelper" "$appDir""/Contents/Library/LaunchServices/com.vmware.fusion.InstallHelper" "$appDir""/Contents/Library/LaunchServices/com.vmware.fusion.CDSHelper" "$appDir""/Contents/Library/LaunchServices/com.vmware.IDHelper" "$appDir""/Contents/Library/Initialize VMware Fusion.tool" )
   if [ -e "$appDir"'/Contents/Resources/Fusion Test Kit.txt' ]; then
      # The substitution of TEST_EXECS uses "$appDir".
      execs+=("$appDir""/Contents/Tests/grainGun" "$appDir""/Contents/Tests/diskTool" "$appDir""/Contents/Tests/diskCreate" "$appDir""/Contents/Tests/batteryTool" "$appDir""/Contents/Tests/dmgTool" "$appDir""/Contents/Tests/smcTool" "$appDir""/Contents/Tests/sslTool" "$appDir""/Contents/Tests/snapshotTool" "$appDir""/Contents/Tests/powerOnTest" "$appDir""/Contents/Tests/nfcTest" "$appDir""/Contents/Tests/pyVigor.zip" )
   fi
   chmod -- "$xAccess" "${execs[@]}"

   # Finally mark some files as set-UID root executables.
   for path in "$@"; do
      # chown/chmod commands may return 0 even on failure (for example
      # on encrypted volumes which are mounted by default with options
      # "nosuid, noowners"). So check again after invoking them.
      chown -- root:"$gName" "$path"
      chmod -- "$sxAccess" "$path"
      isSetuidRoot "$path"
   done
}
perms() {
   local path

   for path in "$@"; do
      if ! isSetuidRoot "$path"; then
         canSet

         setPerms "$@"
         break
      fi
   done
}
# The substitution of MAIN_SETUID_EXECS uses "$appDir".
setuidExecs=("$appDir""/Contents/Library/vmware-vmx" "$appDir""/Contents/Library/vmware-vmx-debug" "$appDir""/Contents/Library/vmware-vmx-stats" "$appDir""/Contents/Library/vmware-authd" "$appDir""/Contents/Library/vmware-rawdiskCreator" "$appDir""/Contents/Library/services/Open VMware Fusion Services" )
if [ -e "$appDir"'/Contents/Resources/Fusion Test Kit.txt' ]; then
   # The substitution of TEST_SETUID_EXECS uses "$appDir".
   setuidExecs+=()
fi

verifyEnvPath() {
   PATHSD_DIR="/etc/paths.d"
   if ! [ -e "$PATHSD_DIR" ]; then
      checkDir "$PATHSD_DIR" 755 0 wheel
   fi
   if [ -d "$PATHSD_DIR" ]; then
      pathFile="$PATHSD_DIR"/com.vmware.fusion.public
      publicDir="$appDir"/Contents/Public
      value=$(<$pathFile)
      [ "$publicDir" = "$value" ]
   fi
}

perms "${setuidExecs[@]}"

if [ "$action" = test ]; then
   verifyEnvPath
fi

if [ "$licenseKey" != '' ]; then
   canSet

   echo '*** Entering serial number unattended...'
   # The license tool returns 0 even on failure. So look at its output.
   "$libDir"/licenses/vmware-licenseTool \
      enter "$licenseKey" '' '' '12.0' 'VMware Fusion for Mac OS' '' \
      | grep -q '200 License operation succeeded.'
fi

if canSet; then
   # Add an entry for CLI/REST/VKD API to be included in PATH.
   # If paths.d does not exists, it should be created.
   PATHSD_DIR="/etc/paths.d"
   if ! [ -e "$PATHSD_DIR" ]; then
      checkDir "$PATHSD_DIR" 755 0 wheel
   fi

   if [ -d "$PATHSD_DIR" ]; then
      # If you move the path,
      # you must update /bora/apps/vmplayer/macos/services/other/services.sh.
      pathFile="$PATHSD_DIR"/com.vmware.fusion.public
      publicDir="$appDir"/Contents/Public
      rm -f "$pathFile".tmp
      print_var "$publicDir" >"$pathFile".tmp
      chmod 444 "$pathFile".tmp
      mv -f "$pathFile".tmp "$pathFile"
   fi

   if [ "$guiPid" != '' ]; then
      # Sleep 5s before killing Fusion to make Finder correctly verify the
      # application bundle.
      OS_MAJOR_VERSION=`sw_vers -productVersion | awk -F. '{print $2}'`
      if [ "$OS_MAJOR_VERSION" -gt "12" ]; then
         sleep 5
      fi

      # Get the associated user from the given pid before killing GUI
      user="`ps -p "$guiPid" -o user=`"

      # Use -9 so Mac OS does not popup a "VMware Fusion quit unexpectedly."
      # message.
      kill -9 -- "$guiPid"
      # Signal delivery is asynchronous. Wait until the process has effectively
      # died.
      while kill -0 -- "$guiPid" 2>/dev/null; do
         true
      done

      xattr -dr com.apple.quarantine "$appDir"
      sudo -u "$user" open -- "$appDir"
   fi

   # Atomically write the location of the last bundle used.
   rm -f "$LOCATION".tmp
   print_var "$appDir" >"$LOCATION".tmp
   chmod 444 "$LOCATION".tmp
   mv -f "$LOCATION".tmp "$LOCATION"

fi
