#!/bin/bash -p
#
# Support program for VMware Fusion.
#
# Collects various configuration and log files
# for use when troubleshooting VMware Fusion.
#


# Print a variable, even if its value is -n (unlike echo).
print_var() {
   printf -- %s "$1"
}


# Canonicalize the path of an existing directory.
canonicalize_dir() {
   # Bash's cd builtin can write to stdout.
   (cd -- "$1" >/dev/null && pwd -P)
}


# Canonicalize the path of an existing file.
canonicalize_file() {
   local file="$1"

   while [ -h "$file" ]; do
      file="`readlink -- "$file"`"
   done

   local dir="`dirname -- "$file"`"
   # 'dir' is a directory, or a symlink to a directory. Canonicalize it.
   print_var "`canonicalize_dir "$dir"`"/"`basename -- "$file"`"
}


# Print how to use this executable, then terminate with an error.
usage() {
   banner \
"Usage: $0" \
"" \
"   -a" \
"      Enforce that this process is invoked with administrative privileges." \
"      Output the ID of this process." \
"      Output the path of the generated file." \
"" \
"   -c <process ID>" \
"      Enforce that this process is invoked with administrative privileges." \
"      Terminate the process whose ID is <process ID>." \
"" \
"   -h" \
"      Print this usage statement." \
"" \
"   -u <user name>" \
"      Assume that this process is invoked with administrative privileges." \
"      Set the name of the actual user to <user name>."
   exit 1
}


# Print any number of strings padded with newlines before and after.
banner() {
   local option

   echo
   for option in "$@"; do
      echo "$option"
   done
   echo
}


# If not in autorun mode, rotate the spinner by one step.
UpdateSpinner() {
   if [ "$AUTORUN" ]; then
      return
   fi

   case "$SPINNER" in
      "|")  SPINNER="/";;
      "/")  SPINNER="-";;
      "-")  SPINNER="\\";;
      "\\") SPINNER="|";;
      *)    SPINNER="|";;
   esac

   echo -en "\rPreparing Files: $SPINNER"
}


# Check for a self contained output directory for later archiving and creates
# it if needed.
check_output_directory() {
   local dir="$1"

   if [ ! -d "$OUTPUT_DIR$dir" ]; then
      mkdir -p "$OUTPUT_DIR$dir"

      if [ $? != 0 ]; then
         banner "Could not create $OUTPUT_DIR$dir... " \
                "Have you run out of disk space?" "Continuing"
         return -1
      fi
   fi
   return 0
}


# Copy whatever files and directories you give it to
# a self contained output directory for later archiving
# Working on copies could slow this down with VERY large files but:
# 1) We don't expect VERY large files
# 2) Copying file first makes sure the archive is coherent
addArchive() {
   local file
   for file in "$@"; do
      local dir=`dirname "$file"`
      check_output_directory "$dir"
      if [ $? != 0 ]; then
         return
      fi

      # Ignore stdout and handle errors.
      UpdateSpinner
      cp -pR "$file" "$OUTPUT_DIR$dir" 2>/dev/null
   done
}


# Execute a command with:
#    stdout redirected to a file to archive.
#    stderr discarded.
runcmd() {
   local outFileRelPath="$1"
   shift # The command arguments are in "$@".

   local dir=`dirname "$outFileRelPath"`
   check_output_directory "$dir"
   if [ $? != 0 ]; then
      return
   fi

   "$@" >"$OUTPUT_DIR$outFileRelPath" 2>/dev/null
   if [ $? != 0 ]; then
      rm "$OUTPUT_DIR$outFileRelPath" 2>/dev/null
      banner "Either could not run" "$@" \
             "or could not write to $OUTPUT_DIR$outFileRelPath" \
             "Do you have a full disk?" "Continuing..."
   fi

   UpdateSpinner
}


# List (optionnally recursively) the contents of a directory, with:
#    stdout redirected to a file to archive.
#    stderr discarded.
runls() {
   local dir="$1"
   local options="$2"

   if [ ! -d "$dir" ]; then
      return
   fi

   runcmd "$dir/ls.txt" ls -l"$options"@Oe -- "$dir"
}


# Call addArchive() for files less than the size given by $1, with the
# file(s) specified by the remaining arguments.
addarchive_smallfiles() {
   local maxSize="$1"
   shift # The files are in "$@".

   local file
   for file in "$@"; do
      local size=`stat -f "%z" "$file" 2>/dev/null`
      if [[ $? != 0 || -z "$size" ]]; then
         continue
      fi

      if [ "$size" -lt "$maxSize" ]; then
         addArchive "$file"
      fi
   done
}


# Parse and copy the descriptor info from monolithic
# sparse disks, and copy other small VMDK files from the given directory.
# This should include all the descriptor info for all disks and snapshots,
# and the partition table files (*-pt.vmdk) for raw disks (Boot Camp, etc).
addarchive_vmdks() {
   local file
   for file in "$@"; do
      local header=`hexdump -n 8 -e '"%x"' "$file" 2>/dev/null`
      if [ "$header" == '564d444b1' ]; then
         # The file header matches the sparse VMDK format version 1
         # Try to locate and extract the descriptor information.
         #
         # The offset and size are 64-bit values representing 512-byte
         # sectors within the file. Hexdump doesn't support 64-bit values,
         # but the descriptor info should never be that far into the file,
         # so we can just use the first 32 bits if the other 32 bits are 0.
         local offset=`hexdump -n 4 -s 28 -e '"%u"' "$file"`
         local offset2=`hexdump -n 4 -s 32 -e '"%u"' "$file"`
         local size=`hexdump -n 4 -s 36 -e '"%u"' "$file"`
         local size2=`hexdump -n 4 -s 40 -e '"%u"' "$file"`
         if [[    offset -gt 0 \
               && offset2 -eq 0 \
               && size -gt 0 \
               && size -lt 200 \
               && size2 -eq 0 ]]; then
            local dir=`dirname "$file"`
            check_output_directory "$dir"
            if [ $? != 0 ]; then
               continue
            fi

            local outFile="$OUTPUT_DIR$file"
            echo "# Descriptor information from sparse VMDK file:" >>"$outFile"
            echo >>"$outFile"
            dd bs=512 if="$file" iseek="$offset" count="$size" \
               >>"$outFile" 2>/dev/null
         fi
      else
         # Add the whole file if it's small. 100K should be enough for
         # regular descriptor files, and most partition table copies.
         addarchive_smallfiles 100000 "$file"
      fi
   done
}


# Archive a VM, i.e. all files of interest in a VM's directory.
addvm() {
   local vmDir="$1"

   addArchive "$vmDir"/*.log
   addArchive "$vmDir"/*.vmx
   addArchive "$vmDir"/*.vmxf
   addArchive "$vmDir"/*.cfg
   addArchive "$vmDir"/vmmcores*
   addArchive "$vmDir"/core
   addArchive "$vmDir"/core.*
   addArchive "$vmDir"/stats
   addArchive "$vmDir"/*.vmsd
   addArchive "$vmDir"/*.vmpl
   # 100K should be enough for powered off snapshots
   addarchive_smallfiles 100000 "$vmDir"/*.vmsn
   addarchive_vmdks "$vmDir"/*.vmdk
   runls "$vmDir" R
}


# Get information about all physical disks and add it to the support archive.
add_diskinfo() {
   # List of all disks and partitions.
   runcmd "/tmp/diskutil-list.txt" diskutil list

   # GPT and MBR info for each disk (along with any stderr output).
   local gptFile="$OUTPUT_DIR/tmp/gpt.txt"
   local fdiskFile="$OUTPUT_DIR/tmp/fdisk.txt"
   local rawdiskFile="$OUTPUT_DIR/tmp/rawdiskCreator.txt"
   local infoFile="$OUTPUT_DIR/tmp/diskutil-info.txt"
   for disk in `ls -1 /dev/disk* | egrep 'disk[0-9]+$'`; do
      echo "$disk" >>"$gptFile"
      gpt -r show "$disk" >>"$gptFile" 2>&1
      echo >>"$gptFile"

      echo "$disk" >>"$fdiskFile"
      fdisk "$disk" >>"$fdiskFile" 2>&1
      echo >>"$fdiskFile"

      echo "$disk" >>"$rawdiskFile"
      "$RAWDISK_CREATOR" print "$disk" >>"$rawdiskFile" 2>&1
      echo >>"$rawdiskFile"

      echo "$disk" >>"$infoFile"
      diskutil info "$disk" >>"$infoFile" 2>&1
      echo >>"$infoFile"
   done
}


# Find all VMware-related user or system process crash/diagnostic logs
# in the given directory, and add them to the archive.
add_crashlogs() {
   local file
   for file in "$1"/*; do
      if grep -qs '^Path:.*\(VM\|vm\)ware' "$file"; then
         addArchive "$file"
      fi
   done
}


# Archive all VMs referenced in inventory file "$1".
get_inventory() {
   local vm
   grep 'config *= *\"/' "$1" \
   | sed 's/^.*config *= *\"//' \
   | sed 's/\".*$//' \
   | sed 's/\\\\/\\/g' \
   | while read vm; do
      # inventory may contain empty values
      if [ "$vm" != "" -a -f "$vm"  ]; then
         addvm "`dirname "$vm"`"
      fi
   done
}


# Cleanup our temp folder and optionally exit.
cleanup() {
   local exitCode="$1"

   rm -rf "$OUTPUT_DIR"
   if [ $? != 0 ]; then
      banner "$OUTPUT_DIR was not successfully removed." \
             "Please remove manually."
   fi

   if [ "$exitCode" ]; then
      exit "$exitCode"
   fi
}


# This executable may run with root privileges, so hardcode a PATH where
# unprivileged users cannot write.
export PATH=/bin:/sbin:/usr/bin:/usr/sbin

THIS="`canonicalize_file "$0"`"
LIBDIR="`dirname -- "$THIS"`"
RAWDISK_CREATOR="$LIBDIR/vmware-rawdiskCreator"
HOST_DIR='/Library/Preferences/VMware Fusion'
printf -v HOME %s ~ # Was bug 841728

# Parse args
while [ $# -gt 0 ]; do
   case "$1" in
   -a)
      AUTORUN=1
      ;;
   -c)
      CANCEL_PID="$2"
      shift
      ;;
   -h)
      usage
      ;;
   -u)
      DEFUSER="$2"
      shift
      ;;
   *)
      usage
      ;;
   esac
   shift
done

if [ ! -z $DEFUSER ]; then
   printf -v HOME %s $(eval echo ~$DEFUSER)
fi

USER_INVENTORY="$HOME/Library/Application Support/VMware Fusion/vmInventory"
FLEX_INVENTORY="$HOME/Library/Application Support/VMware Fusion/FlexInventory"
SHARED_INVENTORY="/Library/Application Support/VMware/VMware Fusion/Shared/vmInventory"

# Place the file on the user's desktop to increases its visibility.
ARCHIVE="$HOME/Desktop/$(date '+VMware Fusion Problem Report %Y-%m-%d at %H.%M.%S').zip"


# Start message

if [ "$AUTORUN" ]; then
   # Output the ID of this process (in hex).
   printf "AUTO_PROCESSID %08x\n" $$
fi

banner "Support program for VMware Fusion - Build 19436697"

# Make sure we have root privileges
# If we are running as root, and a user is not specified in DEFUSER,
# set DEFUSER to root.

if [ "$EUID" -ne 0 ]; then
   if [ "$AUTORUN" ]; then
      banner "Option -a requires administrative privileges."
      exit 1
   fi

   if [ "$CANCEL_PID" ]; then
      banner "Option -c requires administrative privileges."
      exit 1
   fi

   banner "Running this program requires administrative privileges." \
          "You may be prompted for your password below."
   exec sudo "$0" -u "$USER"
elif [ -z "$DEFUSER" ]; then
   DEFUSER=root
fi

# Cancel running support tool

if [ "$CANCEL_PID" ]; then
   printf "Cancelling process %08x\n" "$CANCEL_PID"
   /bin/kill -9 "$CANCEL_PID"
   exit # Use the status of the last command.
fi

# Create a temporary directory to place the files to archive.
#
# SECURITY (was bug 1142940):
# o mktemp creates a new directory, with a random name, and gives it
#   permissions 700 so only the current user (and root) can access the
#   contents of the directory. This prevents a rogue user from:
#   1) Guessing the name of the directory.
#   2) Performing a symlink attack inside the directory.
#   3) Reading data inside the directory, that only the current user (and root)
#      are supposed to access.
# o The directory is created inside /tmp, so only the current user (and root)
#   can delete the directory. This prevents a rogue user from wholesale
#   replacing the directory after we have created it, with a directory that has
#   more lenient permissions.
OUTPUT_DIR="`mktemp -d /tmp/fusion_XXXXXX`"
if [ $? != 0 ]; then
   banner "Could not create a secure temporary directory. Exiting..."
   exit 1
fi

# Cleanup our temp folder if the process is signalled midway.
trap "cleanup 1" HUP INT QUIT TERM ABRT

get_inventory "$USER_INVENTORY"
get_inventory "$SHARED_INVENTORY"

# Add the boot camp helper VM
addvm "$HOME/Library/Application Support/VMware Fusion/Virtual Machines/Helper/naos-1.0.vmwarevm"

# Collect physical disk information
add_diskinfo

addArchive "/etc/crontab"
addArchive "/etc/daily"
addArchive "/etc/monthly"
addArchive "/etc/weekly"

# Add UI logs and file permissions - Fusion specific as well as VMware generic
addArchive "$HOME/Library/Logs/VMware Fusion"/*.log # Fusion specific
addArchive "$HOME/Library/Logs/VMware Fusion Applications Menu"/*.log # Fusion Applications Menu specific
addArchive "$HOME/.vctl"/*.log # VKD specific
addArchive "$HOME/Library/Logs/VMware"/*.log        # VMware generic for user
addArchive "/Library/Logs/VMware"/*.log             # VMware generic for system
runls "$HOME/Library/Logs/VMware Fusion" R
runls "$HOME/Library/Logs/VMware Fusion Applications Menu" R
runls "$HOME/.vctl"

# Add View offline state
addArchive "$HOME/Library/Application Support/VMware Fusion/offline"/*.dict

# Add services
addArchive "/etc/services"

# VMware kernel extensions use the kernel syslog facility. By default,
# '/etc/syslog.conf' directs 'kern.debug' log messages to file
# o '/var/log/system.log' in Mac OS 10.[45]
addArchive /var/log/system.log
addArchive /var/log/system.log.*.gz
addArchive /var/log/system.log.*.bz2
# o '/var/log/kernel.log' in Mac OS 10.6
addArchive /var/log/kernel.log
addArchive /var/log/kernel.log.*.gz
addArchive /var/log/kernel.log.*.bz2

# Logs from kernel modules no longer appear in /var/log/system.log
# or /var/log/kernel.log in macOS Sierra(10.12). Use the log command-line
# tool to view and filter log messages.

runcmd "/tmp/system.log" log show --start $(date -v-4d +"%Y-%m-%d") \
   --style syslog

addArchive /var/log/vmnat*
addArchive /var/log/vmnet-natd-vmnet8.log

# Collect system-wide information
addArchive "$HOST_DIR"/networking
addArchive "$HOST_DIR"/vmnet*/*.conf
addArchive "$HOST_DIR"/config
addArchive "$HOST_DIR"/lastLocationUsed
runls "$HOST_DIR" R
runls "/Library/Application Support/VMware" R
addArchive "/Library/Application Support/VMware/usbarb.rules"
addArchive "/Library/Logs/VMware Fusion Services.log"

# Collect kernel panic data:
# o from Mac OS 10.4:
addArchive "/Library/Logs"/panic.log
# o from Mac OS 10.5:
addArchive "/Library/Logs/PanicReporter"/*.panic
# o from Mac OS 10.6 and newer:
addArchive "/Library/Logs/DiagnosticReports"/Kernel_*.panic
addArchive "/Library/Logs/DiagnosticReports"/Kernel_*.gpuRestart

# Collect system process crashes:
# o from Mac OS 10.5 and earlier:
add_crashlogs "/Library/Logs/CrashReporter/"
# o from Mac OS 10.6 and newer:
add_crashlogs "/Library/Logs/DiagnosticReports/"

# Collect user process crashes:
# o from Mac OS 10.5 and earlier:
add_crashlogs "$HOME/Library/Logs/CrashReporter/"
# o from Mac OS 10.6 and newer:
add_crashlogs "$HOME/Library/Logs/DiagnosticReports/"

# Collect local information and file permissions.
addArchive "$HOME/Library/Preferences/VMware Fusion"
addArchive "$HOME/Library/Preferences/"com.vmware.*.plist
addArchive "$HOME/Library/Logs/VMDKMounter"/*.log
addArchive "$USER_INVENTORY"
addArchive "$SHARED_INVENTORY"
addArchive "$FLEX_INVENTORY"
runls "$HOME/Library/Application Support/VMware Fusion" R
runls "$HOME/Library/Preferences/VMware Fusion" R
# runls only works on one directory, not on a list of files.
runcmd "$HOME/Library/Preferences/prefs_ls.txt" ls -lR@Oe -- \
   "$HOME/Library/Preferences"/com.vmware.*.plist

# Output of various commands.

runcmd "/tmp/vm-support-version.txt" echo "vm-support build: 19436697"
runcmd "/tmp/kexts.txt" /usr/sbin/kextstat
runcmd "/tmp/ioreg-i.txt" ioreg -i
runcmd "/tmp/uname.txt" uname -a
runcmd "/tmp/df.txt" df
runcmd "/tmp/ifconfig.txt" ifconfig -a
runcmd "/tmp/netstat-lan.txt" netstat -lan
runcmd "/tmp/route.txt" netstat -rn
runcmd "/tmp/mount.txt" mount
runcmd "/tmp/dmesg.txt" dmesg
runcmd "/tmp/vm_stat.txt" vm_stat
runcmd "/tmp/uptime.txt" uptime
runcmd "/tmp/date.txt" date
runcmd "/tmp/ps-auwwx.txt" ps auwwx
runcmd "/tmp/ulimit-a.txt" ulimit -a
runcmd "/tmp/umask.txt" umask
runcmd "/tmp/pmset.txt" pmset -g everything

# System Profiler report with the list of all hardware and software.
runcmd "/tmp/system_profiler.txt" system_profiler
runcmd "/tmp/system_profiler.spx" system_profiler -xml
runls /Applications
runls "$HOME/Documents/Virtual Machines" R
runls "$HOME/Documents/Virtual Machines.localized" R

# Terminate the spinner
echo
echo

# Perform the archiving

# SECURITY (was bug 1142940):
# umask 0077 ensures that the archive is created with permissions 600 so only
# the current user (and root) can access the contents of the archive.
(cd "$OUTPUT_DIR" && umask 0077 && zip -rv "$ARCHIVE" .)
if [ $? != 0 ]; then
   banner "The archive did not successfully complete!" \
      "You may wish to re-run this script after checking disk space, etc."
else
   if [ "$AUTORUN" ]; then
      # Output the path of the generated file.
      echo AUTO_SUCCESS "$ARCHIVE"
   fi

   chown "$DEFUSER" "$ARCHIVE"
   st=`stat -s "$ARCHIVE"`
   size=`eval "$st" && echo "$st_size"`
   supportURL='http://www.vmware.com/support/sr/sr_login.jsp'
   if [ $size -gt 10000000 ]; then
      banner \
         "File: $ARCHIVE" \
         "NOTE: $ARCHIVE is greater than 10 MB." \
         "Please do not attach this file when submitting an incident report." \
         "Please contact VMware support for an ftp site." \
         "To file a support incident, go to $supportURL"
   else
      banner "File $ARCHIVE" \
             "has been placed on your desktop." \
             "Please attach this file when submitting an incident report." \
             "To file a support incident, go to $supportURL"
   fi
fi

# Cleanup our temp folder on success.
trap - HUP INT QUIT TERM ABRT; cleanup
