#!/bin/bash -p
# Copyright 2022 VMware, Inc.  All rights reserved.
#
# This executable starts and stops the VMware Fusion services.
#


# Print a variable, even if its value is -n (unlike echo).
print_var() {
   printf -- %s "$1"
}


# Canonicalize the path of an existing directory.
canonicalize_dir() {
   # Bash's cd builtin can write to stdout.
   (cd -- "$1" >/dev/null && pwd -P)
}


# Canonicalize the path of an existing file.
canonicalize_file() {
   local file="$1"

   while [ -h "$file" ]; do
      file="`readlink -- "$file"`"
   done

   local dir="`dirname -- "$file"`"
   # 'dir' is a directory, or a symlink to a directory. Canonicalize it.
   print_var "`canonicalize_dir "$dir"`"/"`basename -- "$file"`"
}


# Test the ownership of a path.
hasOwnership() {
   local path="$1"
   local uid="$2"
   local gid="$3"

   # Numeric IDs are used because some systems have two names (or more)
   # associated with one number.
   set -- `ls -alnd -- "$path" 2>/dev/null`
   [ "$3" = "$uid" -a "$4" = "$gid" ]
}


# Check whether the Apple VMNET Virtualization API is enabled in the config file
isUsingMacosVmnetVirtApi() {
   on=0
   off=1

   # first check if defined explicitly in the config file
   if grep -E '^[[:blank:]]*vmnet\.macos\.useVirtApi[[:blank:]]*=[[:blank:]]*"?TRUE"?' \
         "$HOST_DIR/config" > /dev/null 2>&1 ; then
      return $on
   fi
   if grep -E '^[[:blank:]]*vmnet\.macos\.useVirtApi[[:blank:]]*=[[:blank:]]*"?FALSE"?' \
         "$HOST_DIR/config" > /dev/null 2>&1 ; then
      return $off
   fi

   # if not defined explicitly, check the MacOS version for determining the
   # default value
   ! shouldLoadKernelExtensions
}


# Check whether we should load our kernel extensions.
shouldLoadKernelExtensions() {
   [ `sysctl -n kern.osrelease | cut -d. -f1` -lt 20 ]
}


# Check whether we should load ThinPrint.
shouldLoadThinPrint() {
   [ `uname -m` == "x86_64" ]
}


# Stage and start a privileged helper in exactly the same fashion as SMJobBless.
startPrivilegedHelper() {
   local helper="$1"

   cp -f -- "$LIBDIR"/LaunchServices/"$helper" "$HELPER_DIR"
   chmod 544 "$HELPER_DIR"/"$helper"
   cp -f -- "$LIBDIR"/LaunchServices/"$helper".plist "$HELPERPLIST_DIR"
   chmod 644 "$HELPERPLIST_DIR"/"$helper".plist
   launchctl load "$HELPERPLIST_DIR"/"$helper".plist
}


# Stop and remove a privileged helper.
stopPrivilegedHelper() {
   local helper="$1"

   launchctl stop "$helper" || true
   launchctl unload "$HELPERPLIST_DIR"/"$helper".plist || true
   rm "$HELPER_DIR"/"$helper" || true
   rm  "$HELPERPLIST_DIR"/"$helper".plist || true
}


# Stop vmnet-bridge and remove its related content.
stopVMNetBridge() {
   VMNET_BRIDGE_PID_FILE="/var/run/vmnet-bridge.pid"
   if [ -f "$VMNET_BRIDGE_PID_FILE" ]; then
      vmware_stop_pidfile "$VMNET_BRIDGE_PID_FILE" || true
      rm "$VMNET_BRIDGE_PID_FILE" || true
   fi
}


# Start vmnet services
startVMNetServices() {
   local numRetries=2
   local index=0
   local retString=''
   while [[ $index -lt $numRetries ]]; do
      if retString=`"$LIBDIR/vmnet-cli" --start`; then
         return 0
      fi
      sleep 3
      ((index++))
   done
   echo "$retString"
   return 1
}

# Stop all daemons and helpers.
stopDaemons() {
   "$LIBDIR/vmware-usbarbitrator" --kill || true
   "$LIBDIR/vmnet-cli" --stop || true
   stopVMNetBridge

   stopPrivilegedHelper com.vmware.KextControlHelper
   stopPrivilegedHelper com.vmware.MountHelper
   stopPrivilegedHelper com.vmware.VMMonHelper
   stopPrivilegedHelper com.vmware.DiskHelper
   stopPrivilegedHelper com.vmware.fusion.InstallHelper
   stopPrivilegedHelper com.vmware.IDHelper

   if shouldLoadThinPrint ; then
      # Uninstall ThinPrint.
      bash -p -- "$TP_DIR"/thnusetup.sh -u "$TP_DIR" || true
      rm -rf -- "$TP_DIR" || true
   fi
}

# Load all kexts.
loadKexts() {
   # A kext can only be loaded if it has root:wheel ownership.
   if ! hasOwnership "$LIBDIR"/kexts 0 0; then
      KEXT_DIR="$HOST_DIR"/kexts
      rm -rf -- "$KEXT_DIR"
      cp -r -- "$LIBDIR"/kexts "$KEXT_DIR"
      chown -R -- root:wheel "$KEXT_DIR"
   else
      KEXT_DIR="$LIBDIR"/kexts
   fi

   if ! isUsingMacosVmnetVirtApi ; then
      kextload -r "$KEXT_DIR" -- \
         "$KEXT_DIR"/vmnet.kext \
         "$KEXT_DIR"/vmmon.kext \
         ;
   else
      kextload -r "$KEXT_DIR" -- \
         "$KEXT_DIR"/vmmon.kext \
         ;
   fi

   if ! hasOwnership "$LIBDIR"/kexts 0 0; then
      rm -rf -- "$KEXT_DIR"
   fi
}


# Unload all kexts.
unloadKexts() {
   # It is better if kexts are unloaded in the reverse order they were loaded.
   # See bug 688142.
   if ! isUsingMacosVmnetVirtApi ; then
      kextunload -- \
         "$LIBDIR/kexts/vmmon.kext" \
         "$LIBDIR/kexts/vmnet.kext" \
         || true
   else
      kextunload -- \
         "$LIBDIR/kexts/vmmon.kext" \
         || true
   fi
}


# To debug, uncomment the next 2 lines.
#set -x
#exec >>/tmp/debug 2>&1

set -e
# This executable runs with root privileges, so hardcode a PATH where
# unprivileged users cannot write.
export PATH=/bin:/sbin:/usr/bin:/usr/sbin

# We might have inherited the umask of the calling user. This executable runs
# with root privileges, so hardcode a reasonable and deterministic umask.
# Was bug 923795.
umask 022

THIS="`canonicalize_file "$0"`"
SERVICE_DIR="`dirname -- "$THIS"`"
LIBDIR="`dirname -- "$SERVICE_DIR"`"
CONTENTS_DIR="`dirname -- "$LIBDIR"`"
APP_DIR="`dirname -- "$CONTENTS_DIR"`"
HOST_DIR='/Library/Preferences/VMware Fusion'
TP_DIR="$HOST_DIR"/thnuclnt
LEASEDIR='/var/db/vmware'
HELPER_DIR='/Library/PrivilegedHelperTools'
HELPERPLIST_DIR='/Library/LaunchDaemons'


# BEGINNING_OF_UTIL_DOT_SH
#!/bin/sh
#
# Copyright 2005-2021 VMware, Inc.  All rights reserved.
#
# A few utility functions used by our shell scripts.  Some expect the settings
# database to already be loaded and evaluated.

vmblockmntpt="/proc/fs/vmblock/mountPoint"
vmblockfusemntpt="/var/run/vmblock-fuse"

have_vgauth=@@HAVE_VGAUTH@@
have_caf=@@HAVE_CAF@@

vmware_warn_failure() {
  if [ "`type -t 'echo_warning' 2>/dev/null`" = 'function' ]; then
    echo_warning
  else
    echo -n "$rc_failed"
  fi
}

vmware_failed() {
  if [ "`type -t 'echo_failure' 2>/dev/null`" = 'function' ]; then
    echo_failure
  else
    echo -n "$rc_failed"
  fi
}

vmware_success() {
  if [ "`type -t 'echo_success' 2>/dev/null`" = 'function' ]; then
    echo_success
  else
    echo -n "$rc_done"
  fi
}

# Execute a macro
vmware_exec() {
  local msg="$1"  # IN
  local func="$2" # IN
  shift 2

  echo -n '   '"$msg"

  # On Caldera 2.2, SIGHUP is sent to all our children when this script exits
  # I wanted to use shopt -u huponexit instead but their bash version
  # 1.14.7(1) is too old
  #
  # Ksh does not recognize the SIG prefix in front of a signal name
  if [ "$VMWARE_DEBUG" = 'yes' ]; then
    (trap '' HUP; "$func" "$@")
  else
    (trap '' HUP; "$func" "$@") >/dev/null 2>&1
  fi
  if [ "$?" -gt 0 ]; then
    vmware_failed
    echo
    return 1
  fi

  vmware_success
  echo
  return 0
}

# Execute a macro in the background
vmware_bg_exec() {
  local msg="$1"  # IN
  local func="$2" # IN
  shift 2

  if [ "$VMWARE_DEBUG" = 'yes' ]; then
    # Force synchronism when debugging
    vmware_exec "$msg" "$func" "$@"
  else
    echo -n '   '"$msg"' (background)'

    # On Caldera 2.2, SIGHUP is sent to all our children when this script exits
    # I wanted to use shopt -u huponexit instead but their bash version
    # 1.14.7(1) is too old
    #
    # Ksh does not recognize the SIG prefix in front of a signal name
    (trap '' HUP; "$func" "$@") 2>&1 | logger -t 'VMware-init' -p daemon.err &

    vmware_success
    echo
    return 0
  fi
}

# This is a function in case a future product name contains language-specific
# escape characters.
vmware_product_name() {
  echo '@@PRODUCT_NAME@@'
  exit 0
}

# This is a function in case a future product contains language-specific
# escape characters.
vmware_product() {
  echo '@@PRODUCT@@'
  exit 0
}

is_dsp()
{
   # This is the current way of indicating it is part of a
   # distribution-specific install.  Currently only applies to Tools.
   [ -e "$vmdb_answer_LIBDIR"/dsp ]
}

# They are a lot of small utility programs to create temporary files in a
# secure way, but none of them is standard. So I wrote this
make_tmp_dir() {
  local dirname="$1" # OUT
  local prefix="$2"  # IN
  local tmp
  local serial
  local loop

  tmp="${TMPDIR:-/tmp}"

  # Don't overwrite existing user data
  # -> Create a directory with a name that didn't exist before
  #
  # This may never succeed (if we are racing with a malicious process), but at
  # least it is secure
  serial=0
  loop='yes'
  while [ "$loop" = 'yes' ]; do
    # Check the validity of the temporary directory. We do this in the loop
    # because it can change over time
    if [ ! -d "$tmp" ]; then
      echo 'Error: "'"$tmp"'" is not a directory.'
      echo
      exit 1
    fi
    if [ ! -w "$tmp" -o ! -x "$tmp" ]; then
      echo 'Error: "'"$tmp"'" should be writable and executable.'
      echo
      exit 1
    fi

    # Be secure
    # -> Don't give write access to other users (so that they can not use this
    # directory to launch a symlink attack)
    if mkdir -m 0755 "$tmp"'/'"$prefix$serial" >/dev/null 2>&1; then
      loop='no'
    else
      serial=`expr $serial + 1`
      serial_mod=`expr $serial % 200`
      if [ "$serial_mod" = '0' ]; then
        echo 'Warning: The "'"$tmp"'" directory may be under attack.'
        echo
      fi
    fi
  done

  eval "$dirname"'="$tmp"'"'"'/'"'"'"$prefix$serial"'
}

# Removes "stale" device node
# On udev-based systems, this is never needed.
# On older systems, after an unclean shutdown, we might end up with
# a stale device node while the kernel driver has a new major/minor.
vmware_rm_stale_node() {
   local node="$1"  # IN
   if [ -e "/dev/$node" -a "$node" != "" ]; then
      local node_major=`ls -l "/dev/$node" | awk '{print \$5}' | sed -e s/,//`
      local node_minor=`ls -l "/dev/$node" | awk '{print \$6}'`
      if [ "$node_major" = "10" ]; then
         local real_minor=`cat /proc/misc | grep "$node" | awk '{print \$1}'`
         if [ "$node_minor" != "$real_minor" ]; then
            rm -f "/dev/$node"
         fi
      else
         local node_name=`echo $node | sed -e s/[0-9]*$//`
         local real_major=`cat /proc/devices | grep "$node_name" | awk '{print \$1}'`
         if [ "$node_major" != "$real_major" ]; then
            rm -f "/dev/$node"
         fi
      fi
   fi
}

# Checks if the given pid represents a live process.  If an optional command
# name is provided as argument #2, check that process PID is running that
# command.  Returns 0 if the pid is a live process and the optional command
# name matches, 1 otherwise.
vmware_is_process_alive() {
   local pid="$1" # IN

   if [ $# -gt 1 ];  then
      # Matching complete command name which must be immediately followed
      # by a space or NL.  Older greps will not handle all POSIX regular
      # expressions such as "[ /]vmtoolsd\( \|$\)".
      { ps -f -p $pid | grep $pid | grep  "[/ ]${2}$" > /dev/null 2>&1; } || \
          { ps -f -p $pid | grep $pid | grep  "[/ ]${2} " > /dev/null 2>&1; }
   else
      ps -p $pid | grep $pid > /dev/null 2>&1
   fi
}

# Check if the process associated with a pidfile is running and, if the
# optionally specified command is present, that the process is running
# that command.  Returns 0 if the process matches the specified criteria,
# 1 otherwise.
vmware_check_pidfile() {
   local pidfile="$1" # IN
   local cmd=         # OPT command name
   local pid

   [ $# -gt 1 ] && cmd="$2"
   pid=`cat "$pidfile" 2>/dev/null`
   if [ "$pid" = '' ]; then
      # The file probably does not exist or is empty. Failure
      return 1
   fi
   # Keep only the first number we find, because some Samba pid files are
   # really trashy: they end with NUL characters
   # There is no double quote around $pid on purpose
   set -- $pid
   pid="$1"

   vmware_is_process_alive $pid $cmd
}

# Note:
#  . Each daemon must be started from its own directory to avoid busy devices
#  . Each PID file doesn't need to be added to the installer database, because
#    it is going to be automatically removed when it becomes stale (after a
#    reboot). It must go directly under /var/run, or some distributions
#    (RedHat 6.0) won't clean it
#

# Terminate a process synchronously
vmware_synchrone_kill() {
   local pid="$1"    # IN
   local signal="$2" # IN
   local cmd=        # OPT command name
   local second

   [ $# -gt 2 ] && cmd="$3"
   vmware_is_process_alive $pid $cmd
   if [ "$?" -ne 0 ]; then
      # The service is not running, the following 'kill' command could
      # kill another process if the PID has been reused.
      return 0
   fi

   kill -"$signal" "$pid"

   # Wait a bit to see if the dirty job has really been done
   for second in 0 1 2 3 4 5 6 7 8 9 10; do
      vmware_is_process_alive "$pid" $cmd
      if [ "$?" -ne 0 ]; then
         # Success
         return 0
      fi

      sleep 1
   done

   # Timeout
   return 1
}

# Kill the process associated with a pidfile provided that it is also
# running the expected command if one is specified.
vmware_stop_pidfile() {
   local pidfile="$1" # IN
   local cmd=         # OPT command name
   local pid

   [ $# -gt 1 ] && cmd="$2"
   if [ ! -r "$pidfile" ]; then
      # Missing file. Success
      return 0
   fi

   pid=`cat "$pidfile" 2>/dev/null`
   if [ "$pid" = '' ]; then
      # Empty file. Success
      return 0
   fi
   # Keep only the first number we find, because some Samba pid files are
   # really trashy: they end with NUL characters
   # There is no double quote around $pid on purpose
   set -- $pid
   pid="$1"

   # First try a nice SIGTERM
   if vmware_synchrone_kill "$pid" 15 $cmd; then
      return 0
   fi

   # Then send a strong SIGKILL
   if vmware_synchrone_kill "$pid" 9 $cmd; then
      return 0
   fi

   return 1
}

# Determine if SELinux is enabled
isSELinuxEnabled() {
   if [ "`getenforce 2> /dev/null`" = "Enforcing" ]; then
      echo "yes"
   else
      echo "no"
   fi
}

# Runs a command normally if the SELinux is not enforced.
# Runs a command under the provided SELinux context if the context is passed.
# Runs a command under the parent SELinux context first, then retry under
# the unconfined context if no context is passed.
vmware_exec_selinux() {
   local command="$1"
   local context="$2"

   if [ "`isSELinuxEnabled`" = 'no' ]; then
      # ignore the context parameter
      $command
      return $?
   fi

   # selinux is enforcing...
   if [ -z "$context" ]; then
      # context paramter is missing, try use the parent context
      $command
      retval=$?
      if [ $retval -eq 0 ]; then
	 return $retval
      fi
      # use the unconfined context
      context="unconfined_t"
   fi

   runcon -t $context -- $command
   return $?
}

# Start the blocking file system.  This consists of loading the module and
# mounting the file system.
vmware_start_vmblock() {
   mkdir -p -m 1777 /tmp/VMwareDnD

   # Try FUSE first, fall back on in-kernel module.
   vmware_start_vmblock_fuse && return 0

   vmware_exec 'Loading module' vmware_load_module $vmblock
   exitcode=`expr $exitcode + $?`
   # Check to see if the file system is already mounted.
   if grep -q " $vmblockmntpt vmblock " /etc/mtab; then
       # If it is mounted, do nothing
       true;
   else
       # If it's not mounted, mount it
       vmware_exec_selinux "mount -t vmblock none $vmblockmntpt"
   fi
}

# Stop the blocking file system
vmware_stop_vmblock() {
    # Check if the file system is mounted and only unmount if so.
    # Start with FUSE-based version first, then legacy one.
    #
    # Vmblock-fuse dev path could be /var/run/vmblock-fuse,
    # or /run/vmblock-fuse. Bug 758526.
    if grep -q "/run/vmblock-fuse fuse\.vmware-vmblock " /etc/mtab; then
       # if it's mounted, then unmount it
       vmware_exec_selinux "umount $vmblockfusemntpt"
    fi
    if grep -q " $vmblockmntpt vmblock " /etc/mtab; then
       # if it's mounted, then unmount it
       vmware_exec_selinux "umount $vmblockmntpt"
    fi

    # Unload the kernel module
    vmware_unload_module $vmblock
}

# This is necessary to allow udev time to create a device node.  If we don't
# wait then udev will override the permissions we choose when it creates the
# device node after us.
vmware_delay_for_node() {
   local node="$1"
   local delay="$2"

   while [ ! -e $node -a ${delay} -gt 0 ]; do
      delay=`expr $delay - 1`
      sleep 1
   done
}

vmware_real_modname() {
   # modprobe might be old and not understand the --resolve-alias option, or
   # there might not be an alias. In both cases we assume
   # that the module is not upstreamed.
   mod=$1
   mod_alias=$2

   modname=$(/sbin/modprobe --resolve-alias ${mod_alias} 2>/dev/null)
   if [ $? = 0 -a "$modname" != "" ] ; then
        echo $modname
   else
        echo $mod
   fi
}

vmware_is_upstream() {
   modname=$1
   vmware_exec_selinux "$vmdb_answer_LIBDIR/sbin/vmware-modconfig-console \
                           --install-status" | grep -q "${modname}: other"
   if [ $? = 0 ]; then
      echo "yes"
   else
      echo 'no'
   fi
}

# starts after vmci is loaded
vmware_start_vsock() {
  real_vmci=$(vmware_real_modname $vmci $vmci_alias)

  if [ "`isLoaded "$real_vmci"`" = 'no' ]; then
    # vsock depends on vmci
    return 1
  fi

  real_vsock=$(vmware_real_modname $vsock $vsock_alias)

  vmware_load_module $real_vsock
  vmware_rm_stale_node vsock
  # Give udev 5 seconds to create our node
  vmware_delay_for_node "/dev/vsock" 5
  if [ ! -e /dev/vsock ]; then
     local minor=`cat /proc/misc | grep vsock | awk '{print $1}'`
     mknod --mode=666 /dev/vsock c 10 "$minor"
  else
     chmod 666 /dev/vsock
  fi

  return 0
}

# unloads before vmci
vmware_stop_vsock() {
  # Nothing to do if module is upstream
  if [ "`vmware_is_upstream $vsock`" = 'yes' ]; then
    return 0
  fi

  real_vsock=$(vmware_real_modname $vsock $vsock_alias)
  vmware_unload_module $real_vsock
  rm -f /dev/vsock
}

is_ESX_running() {
  if [ ! -f "$vmdb_answer_LIBDIR"/sbin/vmware-checkvm ] ; then
    echo no
    return
  fi
  if "$vmdb_answer_LIBDIR"/sbin/vmware-checkvm -p | grep -q ESX; then
    echo yes
  else
    echo no
  fi
}

#
# Start vmblock only if ESX is not running and the config script
# built/loaded it (kernel is >= 2.4.0 and  product is tools-for-linux).
# Also don't start when in open-vm compat mode
#
is_vmblock_needed() {
  if [ "`is_ESX_running`" = 'yes' -o "$vmdb_answer_OPEN_VM_COMPAT" = 'yes' ]; then
    echo no
  else
    if [ "$vmdb_answer_VMBLOCK_CONFED" = 'yes' -o \
         "$vmdb_answer_VMBLOCK_CONFED_FUSE" = 'yes' ]; then
      echo yes
    else
      echo no
    fi
  fi
}

VMUSR_PATTERN="(vmtoolsd.*vmusr|vmware-user)"

vmware_signal_vmware_user() {
# Signal all running instances of the user daemon.
# Our pattern ensures that we won't touch the system daemon.
   pkill -$1 -f "$VMUSR_PATTERN"
   return 0
}

# A USR1 causes vmware-user to release any references to vmblock or
# /proc/fs/vmblock/mountPoint, allowing vmblock to unload, but vmware-user
# to continue running. This preserves the user context vmware-user is
# running within. We also shutdown rpc connections to release usage of
# vmci/vsocket.  This is not necessary if operating in OPEN_VM_COMPAT mode.
vmware_user_request_release_resources() {
  if [ "$vmdb_answer_OPEN_VM_COMPAT" != "yes" ]; then
    vmware_signal_vmware_user 'USR1'
  fi
}

# A USR2 causes vmware-user to relaunch itself, picking up vmblock anew.
# This preserves the user context vmware-user is running within.
# This is not necessary if operating in OPEN_VM_COMPAT mode.
vmware_restart_vmware_user() {
  if [ "$vmdb_answer_OPEN_VM_COMPAT" != "yes" ]; then
    vmware_signal_vmware_user 'USR2'
  fi
}

# Checks if there an instance of vmware-user process exists in the system.
is_vmware_user_running() {
  if pgrep -f "$VMUSR_PATTERN" > /dev/null 2>&1; then
    echo yes
  else
    echo no
  fi
}

wrap () {
  AMSG="$1"
  while [ `echo $AMSG | wc -c` -gt 75 ] ; do
    AMSG1=`echo $AMSG | sed -e 's/\(.\{1,75\} \).*/\1/' -e 's/  [ 	]*/  /'`
    AMSG=`echo $AMSG | sed -e 's/.\{1,75\} //' -e 's/  [ 	]*/  /'`
    echo "  $AMSG1"
  done
  echo "  $AMSG"
  echo " "
}

#---------------------------------------------------------------------------
#
# load_settings
#
# Load VMware Installer Service settings
#
# Returns:
#    0 on success, otherwise 1.
#
# Side Effects:
#    vmdb_* variables are set.
#---------------------------------------------------------------------------

load_settings() {
  local settings=`$DATABASE/vmis-settings`
  if [ $? -eq 0 ]; then
    eval "$settings"
    return 0
  else
    return 1
  fi
}

#---------------------------------------------------------------------------
#
# launch_binary
#
# Launch a binary with resolved dependencies.
#
# Returns:
#    None.
#
# Side Effects:
#    Process is replaced with the binary if successful,
#    otherwise returns 1.
#---------------------------------------------------------------------------

launch_binary() {
  local component="$1"		# IN: component name
  shift
  local binary="$2"		# IN: binary name
  shift
  local args="$@"		# IN: arguments
  shift

  # Convert -'s in component name to _ and lookup its libdir
  local component=`echo $component | tr '-' '_'`
  local libdir="vmdb_$component_libdir"

  exec "$libdir"'/bin/launcher.sh'		\
       "$libdir"'/lib'				\
       "$libdir"'/bin/'"$binary"		\
       "$libdir"'/libconf' "$args"
  return 1
}
# END_OF_UTIL_DOT_SH


# Main code

case "$1" in
--start)
   if shouldLoadThinPrint ; then
      # Install ThinPrint.
      rm -rf -- "$TP_DIR"
      cp -r -- "$LIBDIR"/thnuclnt "$TP_DIR"
      bash -p -- "$TP_DIR"/thnusetup.sh -i "$TP_DIR"
   fi

   if shouldLoadKernelExtensions ; then
      loadKexts
   fi

   if ! [ -d "$HELPER_DIR" ]; then
      mkdir "$HELPER_DIR"
      # Match permissions from how SMJobBless creates this directory.
      chmod 1755 "$HELPER_DIR"
   fi

   if ! [ -d "$HELPERPLIST_DIR" ]; then
      mkdir "$HELPERPLIST_DIR"
      chmod 0755 "$HELPERPLIST_DIR"
   fi
   stopPrivilegedHelper  com.vmware.fusion.CDSHelper
   startPrivilegedHelper com.vmware.KextControlHelper
   startPrivilegedHelper com.vmware.MountHelper
   startPrivilegedHelper com.vmware.VMMonHelper
   startPrivilegedHelper com.vmware.DiskHelper
   startPrivilegedHelper com.vmware.IDHelper

   # vmnet-cli and the vmnet daemons need the leases directory to exist.
   mkdir -p -m 755 -- "$LEASEDIR"

   # Cleanup work for the vmnet-bridge failed to quit
   stopVMNetBridge

   # Configure networking if the 'networking' file is missing or
   # the product has been upgraded (i.e. newer 'vmnet-cli') or
   # the networking file has been customized.
   net_time=`stat -f %m "$HOST_DIR/networking" 2>/dev/null` || net_time=0
   cli_time=`stat -f %m "$LIBDIR/vmnet-cli"` || cli_time=0

   if [ "$net_time" -ne "$cli_time" ]; then
      "$LIBDIR/vmnet-cli" --configure
      touch -r "$LIBDIR/vmnet-cli" -- "$HOST_DIR/networking"
   fi

   # Make sure DHCP lease files exist.
   touch -- "$LEASEDIR/vmnet-dhcpd-vmnet1.leases" \
            "$LEASEDIR/vmnet-dhcpd-vmnet8.leases"

   if retString=`"$LIBDIR/vmnet-bridge" -d /var/run/vmnet-bridge.pid`; then
      echo "Started bridge services"
   else
      logger -s -t "VMware Fusion 19436697" \
         "Error: Unable to start the bridge services. Error: $retString [$?]"
      exit 1
   fi

   if startVMNetServices ; then
      echo "Started network services"
   else
      logger -s -t "VMware Fusion 19436697" \
         "Error: Unable to start the network services. Error: $retString [$?]"
      exit 1
   fi

   if retString=`"$LIBDIR/vmware-usbarbitrator"`; then
      echo "Started USB arbitration service"
   else
      logger -s -t "VMware Fusion 19436697" \
         "Error: Unable to start the USB arbitration service. Error: $retString [$?]"
      # Do not fail the services startup if usbarbitrator fails to load.
      # See bug 1427385.
      #exit 1
   fi
   ;;

--stopdaemons)
   stopDaemons
   ;;

--stopkexts)
   unloadKexts
   ;;

--stop)
   stopDaemons
   if shouldLoadKernelExtensions ; then
      unloadKexts
   fi
   ;;

*)
   exit 1
esac
