#!/bin/bash -p

# Copyright(c) 2011 VMware, Inc. All rights reserved.
# A script to open an SMB share for some user
# The user is given exclusive access to the folder.
# Usage: addshare.sh USERNAME PASSWORDFILE SHAREPATH SHARENAME ACLNAME
#
# "aclname" is the name of a user who will be given ACL-based rights
# to the folder being shared.
#
# Prerequisites:
#   -- username must not contain spaces
#   -- sharename must not contain spaces
#
# Note: the "password file" should be a short file containing nothing but
# the password for this user (and avoid a trailing newline, please).
# It's passed though a file to avoid exposing it on the command line.
#

###
### Function to exit with an error message.
###
fail() {
   echo $1
   exit 1
}

###
### Function to remove the shared folder and exit with an error message.
###
removefolder() {
   if [ $CREATEDFOLDER -ne 0 ]; then
       rm -rf "$SHAREPATH" || echo Failed to remove folder $SHAREPATH
   fi
   fail "$1"
}

###
### Function to remove the share and shared folder, then exit with an error message.
###
removeshare() {
   dscl localhost -delete /Local/Default/SharePoints/$SHARENAME || echo Failed to remove share $SHARENAME
   removefolder "$1"
}


###
### Script starts here.
###

###
### This executable runs with root privileges, so hardcode a PATH where
### unprivileged users cannot write.
###
export PATH=/bin:/sbin:/usr/bin:/usr/sbin

if [ $# -ne 5 ]; then fail "Usage: $0 USERNAME PASSWORDFILE SHAREPATH SHARENAME ACLNAME"; fi
if [ $UID -ne 0 ]; then fail "Please run $0 as root."; fi

USERNAME=$1
PASSWORDFILE=$2
SHAREPATH=$3
SHARENAME=$4
ACLNAME=$5

CREATEDFOLDER=0

###
### Neither username nor sharename may contain whitespace; sharepath must not be blank.
###

echo $USERNAME | egrep -q '[[:space:]]' && fail "Error: username must not contain spaces or tabs."
echo $SHARENAME | egrep -q '[[:space:]]' && fail "Error: sharename must not contain spaces or tabs."
if [ "$SHAREPATH" == "" ]; then fail "Usage: $0 USERNAME PASSWORDFILE SHAREPATH SHARENAME ACLNAME"; fi

###
### Read the password from the specified file.
###

if [ ! -f "$PASSWORDFILE" ]; then fail "Password file $PASSWORDFILE was not found"; fi
read PASSWORD < "$PASSWORDFILE"

###
### Create the folder to be shared, if necessary. Assign ownership and exclusive access.
###

if [ ! -d "$SHAREPATH" ]; then
    mkdir "$SHAREPATH" || fail "Cannot create folder $SHAREPATH"
    CREATEDFOLDER=1
fi

chown $USERNAME "$SHAREPATH" || removefolder "failed to chown share folder $SHAREPATH for $USERNAME"
chmod 700 "$SHAREPATH" || removefolder "failed to chmod $SHAREPATH for $USERNAME"

chmod +a "$ACLNAME allow file_inherit,directory_inherit,chown,read,write,append" "$SHAREPATH" \
   || removefolder "Unable to set ACL permissions for $ACLNAME on $SHARENAME"

###
### Share the folder.
###

dscl localhost -create /Local/Default/SharePoints/$SHARENAME || removefolder "dscl failed to create share $SHARENAME"
dscl localhost -append /Local/Default/SharePoints/$SHARENAME directory_path "$SHAREPATH" || removeshare "dscl failed to set smb directory_path"
dscl localhost -append /Local/Default/SharePoints/$SHARENAME smb_appendmask 600 || removeshare "dscl failed to set smb_appendmask"
dscl localhost -append /Local/Default/SharePoints/$SHARENAME smb_directorymask 700 || removeshare "dscl failed to set smb_directorymask"
dscl localhost -append /Local/Default/SharePoints/$SHARENAME smb_guestaccess 0 || removeshare "dscl failed to set smb_guestaccess"
dscl localhost -append /Local/Default/SharePoints/$SHARENAME smb_name $SHARENAME || removeshare "dscl failed to set smb_name"
dscl localhost -append /Local/Default/SharePoints/$SHARENAME smb_shared 1 || removeshare "dscl failed to set smb_shared"

###
### Enable Windows Sharing for this user. Assumes general file sharing has been enabled elsewhere.
###

pwpolicy -u $USERNAME -sethashtypes SMB-NT on || removeshare "pwpolicy failed to enable Windows sharing for $USERNAME"

###
### Reset the password for this user; this persists it to the SMB sharing database.
###

/usr/bin/expect -f - <<EOF
spawn /usr/bin/passwd $USERNAME
expect "assword:"
send "$PASSWORD\r"
expect "assword:"
send "$PASSWORD\r"
send "\r"
expect eof
EOF
if [ $? -ne 0 ]; then removeshare "Failed while resetting password for $USERNAME"; fi

#
# On Snow Leopard, synchronize_shares is needed to persist the dscl changes to the SAMBA files.
#

if [ -f /usr/libexec/samba/synchronize-shares ]; then
   /usr/libexec/samba/synchronize-shares || removeshare "Unable to synchronize sharing info to SAMBA"
fi


