#!/bin/bash -p

# Copyright(c) 2011 VMware, Inc. All rights reserved.
# A simple script to create a user account with password.
# Usage: adduser.sh USERNAME FULLNAME PASSWORDFILE
#
# Prerequisites:
#   -- username must not contain spaces
#
# Note: the "password file" should be a short file containing nothing but
# the password for this user (and avoid a trailing newline, please).
# It's passed though a file to avoid exposing it on the command line.
#
# Adding a user is documented under "Adding a User from the Command Line" at
# http://developer.apple.com/library/mac/#documentation/Porting/Conceptual/PortingUnix/additionalfeatures/additionalfeatures.html

###
### A function to print an error message and exit.
###
fail() {
   echo $1
   exit 1
}

###
### A function to clean up if something fails, then exit with an error message.
###
removeuser () {
  rm -rf /Users/$USERNAME || echo "Failed to remove folder /Users/$USERNAME"
  dscl . -delete /Users/$USERNAME || echo "Failed to delete user $USERNAME"
  fail "$1"
}


###
### Script starts here.
###

###
### This executable runs with root privileges, so hardcode a PATH where
### unprivileged users cannot write.
###
export PATH=/bin:/sbin:/usr/bin:/usr/sbin

if [ $# -ne 3 ]; then fail "Usage: $0 USERNAME FULLNAME PASSWORDFILE"; fi
if [ $UID -ne 0 ]; then fail "Please run $0 as root."; fi

USERNAME=$1
FULLNAME=$2
PASSWORDFILE=$3

###
### Username must not contain whitespace, nor be blank.
###

echo $USERNAME | egrep -q '[[:space:]]' && fail "Error: username must not contain spaces or tabs."
if [ "$USERNAME" == "" ]; then fail "Usage: $0 USERNAME FULLNAME PASSWORDFILE"; fi

###
### Extract the password from the password file.
###

if [ ! -f "$PASSWORDFILE" ]; then fail "Password file $PASSWORDFILE was not found"; fi
read PASSWORD < "$PASSWORDFILE"

###
### Find the next available user ID and the ID of the "staff" group.
###

MAXID=$(dscl . -list /Users UniqueID | awk '{print $2}' | sort -ug | tail -1)
USERID=$((MAXID+1))
STAFFGROUP=$(dscacheutil -q group -a name staff | grep gid | awk '{print $2}')

###
### Create a home folder.
###

mkdir /Users/$USERNAME || fail "Unable to create home folder for user $USERNAME"

###
### Create the user account.
###

dscl . -create /Users/$USERNAME || fail "dscl failed to create user /Users/$USERNAME"
dscl . -create /Users/$USERNAME UserShell /bin/bash || removeuser "dscl failed to set user shell for user $USERNAME"
dscl . -create /Users/$USERNAME RealName "$FULLNAME" || removeuser "Unable to set RealName $FULLNAME for /Users/$USERNAME"
dscl . -create /Users/$USERNAME UniqueID $USERID || removeuser "Unable to set user ID $USERID for /Users/$USERNAME"
dscl . -create /Users/$USERNAME PrimaryGroupID $STAFFGROUP || removeuser "Unable to set primary group id '$STAFFGROUP' for /Users/$USERNAME"
dscl . -create /Users/$USERNAME home /Users/$USERNAME || removeuser "Unable to set home directory for /Users/$USERNAME"

###
### Set the password for this user. Not using dscl for this in order to
### keep the password off the command line.
###

/usr/bin/expect -f - <<EOF
spawn /usr/bin/passwd $USERNAME
expect "assword:"
send "$PASSWORD\r"
expect "assword:"
send "$PASSWORD\r"
send "\r"
expect eof
EOF
if [ $? -ne 0 ]; then removeuser "Failed while setting password for $USERNAME"; fi

###
### Assign ownership of the home folder to the new user.
###

chown $USERNAME /Users/$USERNAME || removeuser "failed to chown home folder for $USERNAME"

echo "Created user $USERNAME"


